# decoding methods to convert from byte string -> spec struct.

from .specs cimport HDF5_01_DataHashSpec, \
    HDF5_00_DataHashSpec, \
    NUMPY_10_DataHashSpec, \
    LMDB_30_DataHashSpec, \
    REMOTE_50_DataHashSpec


cdef HDF5_01_DataHashSpec HDF5_01_Parser(str inp):
    cdef str fmt, uid, cksum, dset, dset_idx
    cdef tuple shape_tup
    cdef list shape_list = []
    cdef int dataset_idx_int
    cdef unsigned char i, c, cc
    cdef unsigned char n = len(inp)
    cdef HDF5_01_DataHashSpec res

    c = 0
    cc = 0
    for i in range(n):
        if inp[i] == ':':
            if cc == 0:
                fmt = inp[c:i]
            elif cc == 1:
                uid = inp[c:i]
            elif cc == 2:
                cksum = inp[c:i]
            elif cc == 3:
                dset = inp[c:i]
            elif cc == 4:
                dset_idx = inp[c:i]
            c = i + 1
            cc = cc + 1
    shape_vs = inp[c:n]

    c = 0
    n = len(shape_vs)
    for i in range(n):
        if shape_vs[i] == ' ':
            shape_list.append(int(shape_vs[c:i]))
            c = i + 1
    if shape_vs[c:n] != '':
        shape_list.append(int(shape_vs[c:]))

    shape_tup = tuple(shape_list)
    dataset_idx_int = int(dset_idx)
    res = HDF5_01_DataHashSpec(fmt, uid, cksum, dset, dataset_idx_int, shape_tup)
    return res


cdef HDF5_00_DataHashSpec HDF5_00_Parser(str inp):
    cdef str fmt, uid, cksum, dset, dset_idx
    cdef tuple shape_tup
    cdef list shape_list = []
    cdef int dataset_idx_int
    cdef unsigned char i, c, cc
    cdef unsigned char n = len(inp)
    cdef HDF5_00_DataHashSpec res

    c = 0
    cc = 0
    for i in range(n):
        if inp[i] == ':':
            if cc == 0:
                fmt = inp[c:i]
            elif cc == 1:
                uid = inp[c:i]
            elif cc == 2:
                cksum = inp[c:i]
            elif cc == 3:
                dset = inp[c:i]
            elif cc == 4:
                dset_idx = inp[c:i]
            c = i + 1
            cc = cc + 1
    shape_vs = inp[c:n]

    c = 0
    n = len(shape_vs)
    for i in range(n):
        if shape_vs[i] == ' ':
            shape_list.append(int(shape_vs[c:i]))
            c = i + 1
    if shape_vs[c:n] != '':
        shape_list.append(int(shape_vs[c:]))

    shape_tup = tuple(shape_list)
    dataset_idx_int = int(dset_idx)
    res = HDF5_00_DataHashSpec(fmt, uid, cksum, dset, dataset_idx_int, shape_tup)
    return res


cdef NUMPY_10_DataHashSpec NUMPY_10_Parser(str inp):
    cdef str fmt, uid, cksum, collection_idx
    cdef tuple shape_tup
    cdef list shape_list = []
    cdef int collection_idx_int
    cdef unsigned char i, c, cc
    cdef unsigned char n = len(inp)
    cdef NUMPY_10_DataHashSpec res

    c = 0
    cc = 0
    for i in range(n):
        if inp[i] == ':':
            if cc == 0:
                fmt = inp[c:i]
            elif cc == 1:
                uid = inp[c:i]
            elif cc == 2:
                cksum = inp[c:i]
            elif cc == 3:
                collection_idx = inp[c:i]
            c = i + 1
            cc = cc + 1
    shape_vs = inp[c:n]

    c = 0
    n = len(shape_vs)
    for i in range(n):
        if shape_vs[i] == ' ':
            shape_list.append(int(shape_vs[c:i]))
            c = i + 1
    if shape_vs[c:n] != '':
        shape_list.append(int(shape_vs[c:]))

    shape_tup = tuple(shape_list)
    collection_idx_int = int(collection_idx)
    res = NUMPY_10_DataHashSpec(fmt, uid, cksum, collection_idx_int, shape_tup)
    return res


cdef LMDB_30_DataHashSpec LMDB_30_Parser(str inp):
    cdef str fmt, uid, row_idx, checksum
    cdef unsigned char i, c, cc
    cdef unsigned char n = len(inp)
    cdef LMDB_30_DataHashSpec res

    c = 0
    cc = 0
    for i in range(n):
        if inp[i] == ':':
            if cc == 0:
                fmt = inp[c:i]
            elif cc == 1:
                uid = inp[c:i]
            elif cc == 2:
                row_idx = inp[c:i]
            c = i + 1
            cc = cc + 1
    checksum = inp[c:n]

    res = LMDB_30_DataHashSpec(fmt, uid, row_idx, checksum)
    return res



cdef REMOTE_50_DataHashSpec REMOTE_50_Parser(str inp):
    cdef str fmt, schema_hash
    cdef unsigned char i, c
    cdef unsigned char n = len(inp)
    cdef REMOTE_50_DataHashSpec res

    c = 0
    for i in range(n):
        if inp[i] == ':':
            fmt = inp[c:i]
            c = i + 1
    schema_hash = inp[c:]
    res = REMOTE_50_DataHashSpec(fmt, schema_hash)
    return res


cpdef object backend_decoder(bytes inp):
    cdef str backend, inp_str
    inp_str = inp.decode('utf-8')
    backend = inp_str[:2]
    if backend == '00':
        return HDF5_00_Parser(inp_str)
    elif backend == '01':
        return HDF5_01_Parser(inp_str)
    elif backend == '10':
        return NUMPY_10_Parser(inp_str)
    elif backend == '30':
        return LMDB_30_Parser(inp_str)
    elif backend == '50':
        return REMOTE_50_Parser(inp_str)
    else:
        raise ValueError(f'unknown backend type for input str {inp_str}')



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import os
import platform
import sys
from os.path import join
from distutils.sysconfig import get_config_var
from distutils.version import LooseVersion
from setuptools import setup, Extension, find_packages


# Use `setup.py [] --debug` for a debug build of hangar
HANGAR_DEBUG_BUILD = False


# Set deployment target for mac
#
# Need to ensure that extensions are built for macos 10.9 when compiling on a
# 10.9 system or above, overriding distutils behavior which is to target
# the version used to build the current python binary.
#
# TO OVERRIDE:
#   set MACOSX_DEPLOYMENT_TARGET before calling setup.py
#
# From https://github.com/pandas-dev/pandas/pull/24274
# 3-Clause BSD License: https://github.com/pandas-dev/pandas/blob/master/LICENSE
if sys.platform == 'darwin':
    if 'MACOSX_DEPLOYMENT_TARGET' not in os.environ:
        current_system = LooseVersion(platform.mac_ver()[0])
        python_target = LooseVersion(get_config_var('MACOSX_DEPLOYMENT_TARGET'))
        if python_target < '10.9' and current_system >= '10.9':
            os.environ['MACOSX_DEPLOYMENT_TARGET'] = '10.9'


class LazyCommandClass(dict):
    """
    Lazy command class that defers operations requiring Cython and numpy until
    they've actually been downloaded and installed by setup_requires.
    """

    def __contains__(self, key):
        return key in ['build_ext', 'bdist_wheel', 'sdist'] or super().__contains__(key)

    def __setitem__(self, key, value):
        if key == 'build_ext':
            raise AssertionError("build_ext overridden!")
        super().__setitem__(key, value)

    def __getitem__(self, key):
        if key == 'build_ext':
            return self.make_build_ext_cmd()
        elif key == 'bdist_wheel':
            return self.make_bdist_wheel_cmd()
        elif key  == 'sdist':
            return self.make_sdist_cmd()
        else:
            return super().__getitem__(key)

    def make_build_ext_cmd(self):
        """Returns a command class implementing 'build_ext'.
        """
        from Cython.Distutils.build_ext import new_build_ext as cython_build_ext
        from Cython.Compiler.Main import default_options

        default_options['language_level'] = 3
        default_options['compiler_directives']['embedsignature'] = True
        default_options['compiler_directives']['emit_code_comments'] = True
        if HANGAR_DEBUG_BUILD is True:
            default_options['annotate'] = True
            default_options['emit_linenums'] = True
            default_options['gdb_debug'] = True

        class build_ext(cython_build_ext):
            def build_extensions(self):
                cython_build_ext.build_extensions(self)

        return build_ext

    def make_bdist_wheel_cmd(self):
        """Returns a command class implementing 'bdist_wheel'.
        """
        from wheel.bdist_wheel import bdist_wheel

        class bdist_wheel_cmd(bdist_wheel):
            def run(self):
                # This may modify package_data:
                bdist_wheel.run(self)

        return bdist_wheel_cmd

    def make_sdist_cmd(self):
        """A command class implementing 'sdist'.
        """
        from distutils.command.sdist import sdist as _sdist

        class sdist(_sdist):
            def run(self):
                # Make sure the compiled Cython files in the distribution are up-to-date
                # so we generate .c files correctly (.so will be removed)
                _sdist.run(self)

        return sdist

# Pass command line flags to setup.py script
# handle --lflags=[FLAGS] --cflags=[FLAGS]
args = sys.argv[:]
for arg in args:
    if arg.find('--debug') == 0:
        HANGAR_DEBUG_BUILD = True
        sys.argv.remove(arg)

# Source files for build
CYTHON_SOURCES = [
    join('src', 'hangar', 'optimized_utils.pyx'),
    join('src', 'hangar', 'backends', 'specs.pyx'),
    join('src', 'hangar', 'backends', 'specparse.pyx'),
    join('src', 'hangar', 'records', 'recordstructs.pyx'),
    join('src', 'hangar', 'records', 'column_parsers.pyx'),
    join('src', 'hangar', 'records', 'hashmachine.pyx'),
]
CYTHON_HEADERS = [
    join('src', 'hangar', 'optimized_utils.pxd'),
    join('src', 'hangar', 'backends', 'specs.pxd'),
    join('src', 'hangar', 'records', 'recordstructs.pxd'),
]

__extensions = []
for source in CYTHON_SOURCES:
    module_name = os.path.splitext(source)[0]
    if module_name + '.pxd' in CYTHON_HEADERS:
        deps = module_name + '.pxd'
    else:
        deps = None
    if module_name.startswith(f'src{os.sep}'):
        module_name = module_name.lstrip(f'src{os.sep}')
    module_name = module_name.replace(os.sep, '.')
    ext = Extension(module_name,
                    include_dirs=[],
                    define_macros=[],
                    sources=[source],
                    depends=[deps] if deps else [],
                    library_dirs=[],
                    libraries=[],
                    extra_link_args=[],
                    extra_compile_args=[],
                    language="c")
    __extensions.append(ext)

with open('README.rst') as f:
    README_RST = f.read()

SHORT_DESCRIPTION = (
    'Hangar is version control for tensor data. Commit, branch, merge, '
    'revert, and collaborate in the data-defined software era.'
)

SETUP_REQUIRES = [
    'cython>=0.27',
    'setuptools>=40.0',
    'wheel>=0.30',
]

INSTALL_REQUIRES = [
    'blosc>=1.8',
    'click',
    'grpcio',
    'protobuf',
    'h5py>=2.9',
    'hdf5plugin>=2.0',
    'lmdb>=0.94',
    'numpy',
    'tqdm',
    'wrapt',
    'xxhash',
]

setup(
    name='hangar',
    version='0.5.0',
    license='Apache 2.0',
    # Package Meta Info (for PyPi)
    description=SHORT_DESCRIPTION,
    long_description=README_RST,
    long_description_content_type='text/x-rst',
    author='Richard Izzo',
    author_email='rick@tensorwerk.com',
    maintainer='Richard Izzo',
    maintainer_email='rick@tensorwerk.com',
    url='https://github.com/tensorwerk/hangar-py',
    project_urls={
        'Documentation': 'https://hangar-py.readthedocs.io/',
        'Changelog': 'https://hangar-py.readthedocs.io/en/latest/changelog.html',
        'Issue Tracker': 'https://github.com/tensorwerk/hangar-py/issues',
    },
    platforms=['any'],
    # Module Source Files
    ext_modules=__extensions,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'': ['*.ini', '*.proto']},
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': ['hangar = hangar.cli:main']
    },
    # Requirements
    python_requires='>= 3.6.0',
    install_requires=INSTALL_REQUIRES,
    setup_requires=SETUP_REQUIRES,
    # hooks into `sdist`, `bdist_wheel`, `bdist_ext` commands.
    cmdclass=LazyCommandClass(),
    # PyPi classifiers
    # http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Database',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Version Control',
        'Topic :: Utilities',
    ],
)



import pytest
import numpy as np
from conftest import fixed_shape_backend_params


def assert_equal(arr, arr2):
    assert np.array_equal(arr, arr2)
    assert arr.dtype == arr2.dtype



@pytest.fixture(scope='class')
def subsample_data_map():
    arr = np.arange(5*7).astype(np.uint16).reshape((5, 7))
    res = {
        'foo': {
            0: arr,
            1: arr + 1,
            2: arr + 2
        },
        2: {
            'bar': arr + 3,
            'baz': arr + 4
        }
    }
    return res


@pytest.fixture(scope='class')
def sample_data_map():
    arr = np.arange(5*7).astype(np.uint16).reshape((5, 7))
    res = {
        0: arr,
        1: arr + 1,
        2: arr + 2,
        'bar': arr + 3,
        'baz': arr + 4,
    }
    return res


@pytest.fixture(params=fixed_shape_backend_params, scope='class')
def backend_param(request):
    return request.param


@pytest.fixture(params=[False, True], scope='class')
def write_enabled(request):
    return request.param


@pytest.fixture(params=[False, True], scope='class')
def contains_subsamples(request):
    return request.param


@pytest.fixture(scope='class')
def initialized_column(
    write_enabled, backend_param, contains_subsamples, classrepo, subsample_data_map, sample_data_map
):
    co = classrepo.checkout(write=True)
    aset = co.add_ndarray_column(f'foo{backend_param}{int(write_enabled)}{int(contains_subsamples)}',
                                    shape=(5, 7), dtype=np.uint16,
                                    backend=backend_param, contains_subsamples=contains_subsamples)
    if contains_subsamples:
        aset.update(subsample_data_map)
    else:
        aset.update(sample_data_map)
    co.commit(f'done {backend_param}{write_enabled}{contains_subsamples}')
    co.close()
    if write_enabled:
        nco = classrepo.checkout(write=True)
        yield nco.columns[f'foo{backend_param}{int(write_enabled)}{int(contains_subsamples)}']
        nco.close()
    else:
        nco = classrepo.checkout()
        yield nco.columns[f'foo{backend_param}{int(write_enabled)}{int(contains_subsamples)}']
        nco.close()


@pytest.fixture(scope='class')
def initialized_column_read_only(backend_param, contains_subsamples, classrepo, subsample_data_map, sample_data_map):
    co = classrepo.checkout(write=True)
    aset = co.add_ndarray_column(f'foo{backend_param}{int(contains_subsamples)}',
                                    shape=(5, 7), dtype=np.uint16,
                                    backend=backend_param, contains_subsamples=contains_subsamples)
    if contains_subsamples:
        aset.update(subsample_data_map)
    else:
        aset.update(sample_data_map)

    digest = co.commit(f'done {backend_param}{contains_subsamples}')
    co.close()
    nco = classrepo.checkout(write=False, commit=digest)
    yield nco.columns[f'foo{backend_param}{int(contains_subsamples)}']
    nco.close()


class TestPickleableColumns:

    def test_is_pickleable(self, initialized_column, sample_data_map, subsample_data_map):
        import pickle

        aset = initialized_column
        if aset.iswriteable:
            with pytest.raises(PermissionError, match='Method "__getstate__" cannot'):
                pickle.dumps(aset, protocol=pickle.HIGHEST_PROTOCOL)
        else:
            pkl = pickle.dumps(aset, protocol=pickle.HIGHEST_PROTOCOL)
            assert isinstance(pkl, bytes)


class TestLoadableColumns:

    def test_is_pickle_is_loadable(self, initialized_column_read_only, sample_data_map, subsample_data_map):
        import pickle

        aset = initialized_column_read_only
        pkl = pickle.dumps(aset, protocol=pickle.HIGHEST_PROTOCOL)
        assert isinstance(pkl, bytes)
        equiv = pickle.loads(pkl)

        if aset.contains_subsamples:
            assert len(aset) == len(subsample_data_map)
            assert len(equiv) == len(subsample_data_map)

            for sample_key, subsample_data in subsample_data_map.items():
                assert sample_key in aset
                assert sample_key in equiv
                aset_sample = aset[sample_key]
                equiv_sample = equiv[sample_key]
                assert len(aset_sample) == len(subsample_data)
                assert len(equiv_sample) == len(subsample_data)

                for subsample_key, expected in subsample_data.items():
                    assert subsample_key in aset_sample
                    assert subsample_key in equiv_sample
                    assert_equal(aset_sample[subsample_key], expected)
                    assert_equal(equiv_sample[subsample_key], expected)
        else:
            assert len(aset) == len(sample_data_map)
            assert len(equiv) == len(sample_data_map)
            for sample_key, expected in sample_data_map.items():
                assert sample_key in aset
                assert sample_key in equiv
                assert_equal(aset[sample_key], expected)
                assert_equal(equiv[sample_key], expected)
        equiv._destruct()
        del equiv



import pytest

import numpy as np


param_shapes = [(1,), (1000,), (1, 1), (623, 3, 5), (2, 4, 5, 6, 1, 3)]
param_dtypes = [np.uint8, np.float32, np.float64, np.int32]
param_digest = ['0=digesta', '0=digestaaaaaa', '2=digestaaaaaaaaaaaaaaaaaaaaaaaaaa']
param_schema = ['schemaa', 'schemaaaaaaaaa', 'schemaaaaaaaaaaaaaaaa']


def assert_equal(arr, arr2):
    assert np.array_equal(arr, arr2)
    assert arr.dtype == arr2.dtype


@pytest.fixture(scope='module', params=param_shapes)
def arr_shape(request):
    return request.param

@pytest.fixture(scope='module', params=param_dtypes)
def arr_dtype(request):
    return request.param

@pytest.fixture(scope='module', params=param_digest)
def ident_digest(request):
    return request.param

@pytest.fixture(scope='module', params=param_schema)
def ident_schema(request):
    return request.param


@pytest.fixture(scope='module')
def array_testcase(arr_shape, arr_dtype):
    arr = 200 * np.random.random_sample(arr_shape) - 100
    return arr.astype(arr_dtype)


@pytest.fixture(scope='module')
def ident_testcase(ident_digest, ident_schema):
    return (ident_digest, ident_schema)


def test_serialize_deserialize_array(array_testcase):
    from hangar.remote.chunks import serialize_arr
    from hangar.remote.chunks import deserialize_arr

    raw = serialize_arr(array_testcase)
    res = deserialize_arr(raw)
    assert_equal(array_testcase, res)


def test_serialize_deserialize_ident(ident_testcase):
    from hangar.remote.chunks import serialize_ident
    from hangar.remote.chunks import deserialize_ident
    from hangar.remote.chunks import ArrayIdent

    digest, schema = ident_testcase
    raw = serialize_ident(digest, schema)
    res = deserialize_ident(raw)
    assert isinstance(res, ArrayIdent)
    assert res.digest == digest
    assert res.schema == schema


def test_serialize_deserialize_record(array_testcase, ident_testcase):
    from hangar.remote.chunks import serialize_record
    from hangar.remote.chunks import deserialize_record
    from hangar.remote.chunks import ArrayRecord

    digest, schema = ident_testcase
    raw = serialize_record(array_testcase, digest, schema)
    res = deserialize_record(raw)
    assert isinstance(res, ArrayRecord)
    assert_equal(res.array, array_testcase)
    assert res.digest == digest
    assert res.schema == schema


@pytest.mark.parametrize('nrecords', [1, 25])
def test_serialize_deserialize_record_pack(ident_testcase, nrecords):
    from hangar.remote.chunks import serialize_record
    from hangar.remote.chunks import serialize_record_pack
    from hangar.remote.chunks import deserialize_record
    from hangar.remote.chunks import deserialize_record_pack
    from hangar.remote.chunks import ArrayRecord

    idx = 0
    ArrList, RecList = [], []
    digest, schema = ident_testcase
    for shape in param_shapes:
        for dtype in param_dtypes:
            arr = 200 * np.random.random_sample(shape) + 100
            arr = arr.astype(dtype)
            digest = f'digest{str(idx)*len(digest)}'
            schema = f'schema{str(idx)*len(schema)}'
            idx += 1

            ArrList.append((arr, digest, schema))
            RecList.append(serialize_record(arr, digest, schema))

    rawpack = serialize_record_pack(RecList)
    reslist = deserialize_record_pack(rawpack)

    assert reslist == RecList

    for rawres, origRec in zip(reslist, ArrList):
        resRec = deserialize_record(rawres)
        assert isinstance(resRec, ArrayRecord)
        assert_equal(resRec.array, origRec[0])
        assert resRec.digest == origRec[1]
        assert resRec.schema == origRec[2]


def test_serialize_deserialize_ident_digest_field_only(ident_testcase):
    from hangar.remote.chunks import serialize_ident
    from hangar.remote.chunks import deserialize_ident
    from hangar.remote.chunks import ArrayIdent

    digest, schema = ident_testcase
    raw = serialize_ident(digest, '')
    res = deserialize_ident(raw)
    assert isinstance(res, ArrayIdent)
    assert res.digest == digest
    assert res.schema == ''


def test_serialize_deserialize_ident_schema_field_only(ident_testcase):
    from hangar.remote.chunks import serialize_ident
    from hangar.remote.chunks import deserialize_ident
    from hangar.remote.chunks import ArrayIdent

    digest, schema = ident_testcase
    raw = serialize_ident('', schema)
    res = deserialize_ident(raw)
    assert isinstance(res, ArrayIdent)
    assert res.digest == ''
    assert res.schema == schema


@pytest.mark.parametrize('nrecords', [1, 25])
def test_serialize_deserialize_ident_only_record_pack(ident_testcase, nrecords):
    from hangar.remote.chunks import serialize_ident
    from hangar.remote.chunks import deserialize_ident
    from hangar.remote.chunks import serialize_record_pack
    from hangar.remote.chunks import deserialize_record_pack
    from hangar.remote.chunks import ArrayIdent

    idx = 0
    IdentList, RawList = [], []
    digest, schema = ident_testcase
    for idx in range(nrecords):
        digest = f'digest{str(idx)*len(digest)}'
        schema = f'schema{str(idx)*len(schema)}'

        IdentList.append((digest, schema))
        RawList.append(serialize_ident(digest, schema))

    packed_raw = serialize_record_pack(RawList)
    unpacked_raw = deserialize_record_pack(packed_raw)

    assert unpacked_raw == RawList

    for raw, origIdent in zip(unpacked_raw, IdentList):
        resIdent = deserialize_ident(raw)
        assert isinstance(resIdent, ArrayIdent)
        assert resIdent.digest == origIdent[0]
        assert resIdent.schema == origIdent[1]


@pytest.mark.parametrize('nrecords', [1, 25])
def test_serialize_deserialize_ident_only_digest_only_record_pack(ident_testcase, nrecords):
    from hangar.remote.chunks import serialize_ident
    from hangar.remote.chunks import deserialize_ident
    from hangar.remote.chunks import serialize_record_pack
    from hangar.remote.chunks import deserialize_record_pack
    from hangar.remote.chunks import ArrayIdent

    idx = 0
    IdentList, RawList = [], []
    digest, schema = ident_testcase
    for idx in range(nrecords):
        digest = f'digest{str(idx)*len(digest)}'
        schema = f''

        IdentList.append((digest, schema))
        RawList.append(serialize_ident(digest, schema))

    packed_raw = serialize_record_pack(RawList)
    unpacked_raw = deserialize_record_pack(packed_raw)

    assert unpacked_raw == RawList

    for raw, origIdent in zip(unpacked_raw, IdentList):
        resIdent = deserialize_ident(raw)
        assert isinstance(resIdent, ArrayIdent)
        assert resIdent.digest == origIdent[0]
        assert resIdent.schema == origIdent[1]


@pytest.mark.parametrize('nrecords', [1, 25])
def test_serialize_deserialize_ident_only_schema_only_record_pack(ident_testcase, nrecords):
    from hangar.remote.chunks import serialize_ident
    from hangar.remote.chunks import deserialize_ident
    from hangar.remote.chunks import serialize_record_pack
    from hangar.remote.chunks import deserialize_record_pack
    from hangar.remote.chunks import ArrayIdent

    idx = 0
    IdentList, RawList = [], []
    digest, schema = ident_testcase
    for idx in range(nrecords):
        digest = f''
        schema = f'schema{str(idx)*len(schema)}'

        IdentList.append((digest, schema))
        RawList.append(serialize_ident(digest, schema))

    packed_raw = serialize_record_pack(RawList)
    unpacked_raw = deserialize_record_pack(packed_raw)

    assert unpacked_raw == RawList

    for raw, origIdent in zip(unpacked_raw, IdentList):
        resIdent = deserialize_ident(raw)
        assert isinstance(resIdent, ArrayIdent)
        assert resIdent.digest == origIdent[0]
        assert resIdent.schema == origIdent[1]


import pytest
import numpy as np

from conftest import (
    variable_shape_backend_params,
    fixed_shape_backend_params,
    str_variable_shape_backend_params
)

import string
from hypothesis import given, settings, HealthCheck
import hypothesis.strategies as st
from hypothesis.extra import numpy as npst

from hangar import Repository


# ------------------------ Fixture Setup ------------------------------


added_samples = set()


@pytest.fixture(params=fixed_shape_backend_params, scope='class')
def fixed_shape_repo_co_float32_aset_flat(classrepo, request) -> Repository:
    # needed because fixtures don't reset between each hypothesis run
    # tracks added_samples = set(sample_key)
    global added_samples
    added_samples = set()
    co = classrepo.checkout(write=True)
    co.add_ndarray_column(name='writtenaset',
                          shape=(5, 5, 5),
                          dtype=np.float32,
                          variable_shape=False,
                          backend=request.param,
                          contains_subsamples=False)
    yield co
    co.reset_staging_area()
    co.close()


@pytest.fixture(params=variable_shape_backend_params, scope='class')
def variable_shape_repo_co_float32_aset_flat(classrepo, request) -> Repository:
    # needed because fixtures don't reset between each hypothesis run
    # tracks added_samples = set(sample_key)
    global added_samples
    added_samples = set()
    co = classrepo.checkout(write=True)
    co.add_ndarray_column(name='writtenaset',
                          shape=(5, 5, 5),
                          dtype=np.float32,
                          variable_shape=True,
                          backend=request.param,
                          contains_subsamples=False)
    yield co
    co.reset_staging_area()
    co.close()


@pytest.fixture(params=variable_shape_backend_params, scope='class')
def variable_shape_repo_co_uint8_aset_flat(classrepo, request) -> Repository:
    # needed because fixtures don't reset between each hypothesis run
    # tracks added_samples = set(sample_key)
    global added_samples
    added_samples = set()
    co = classrepo.checkout(write=True)
    co.add_ndarray_column(name='writtenaset',
                          shape=(5, 5, 5),
                          dtype=np.uint8,
                          variable_shape=True,
                          backend=request.param,
                          contains_subsamples=False)
    yield co
    co.reset_staging_area()
    co.close()


@pytest.fixture(params=str_variable_shape_backend_params, scope='class')
def variable_shape_repo_co_str_aset_flat(classrepo, request) -> Repository:
    # needed because fixtures don't reset between each hypothesis run
    # tracks added_samples = set(sample_key)
    global added_samples
    added_samples = set()
    co = classrepo.checkout(write=True)
    co.add_str_column(name='strcolumn',
                      contains_subsamples=False,
                      backend=request.param)
    yield co
    co.reset_staging_area()
    co.close()


# -------------------------- Test Generation ---------------------------------
# Test cases are encapsulated in a classes (and fixture functions are set to
# "class" level scope in order to handle a warning introduced in hypothesis
# version 5.6.0 - 2020-02-29
#
# From release notes:
# > This release adds an explicit warning for tests that are both decorated with @given(...)
#   and request a function-scoped pytest fixture, because such fixtures are only executed once
#   for all Hypothesis test cases and that often causes trouble. See issue #377
#   (https://github.com/HypothesisWorks/hypothesis/issues/377)
#
# However, this is actually the intended behavior for hangar, since we ant to reuse
# the same repo/checkout across all of the test case inputs that hypothesis generates


st_valid_names = st.text(
    min_size=1, max_size=8, alphabet=string.ascii_letters + string.digits + '_-.')
st_valid_ints = st.integers(min_value=0, max_value=999_999)
st_valid_keys = st.one_of(st_valid_ints, st_valid_names)


valid_arrays_fixed = npst.arrays(np.float32,
                                 shape=(5, 5, 5),
                                 fill=st.floats(min_value=-10,
                                                max_value=10,
                                                allow_nan=False,
                                                allow_infinity=False,
                                                width=32),
                                 elements=st.floats(min_value=-10,
                                                    max_value=10,
                                                    allow_nan=False,
                                                    allow_infinity=False,
                                                    width=32))


class TestColumn1:

    @given(key=st_valid_keys, val=valid_arrays_fixed)
    @settings(max_examples=200, deadline=None)
    def test_arrayset_fixed_key_values(self, key, val, fixed_shape_repo_co_float32_aset_flat):
        global added_samples

        co = fixed_shape_repo_co_float32_aset_flat
        col = co.columns['writtenaset']
        assert col.schema_type == 'fixed_shape'

        col[key] = val
        added_samples.add(key)
        out = col[key]
        assert out.dtype == val.dtype
        assert out.shape == val.shape
        assert np.allclose(out, val)
        assert len(col) == len(added_samples)


valid_shapes_var = npst.array_shapes(min_dims=3, max_dims=3, min_side=1, max_side=5)
valid_arrays_var_float32 = npst.arrays(np.float32,
                                       shape=valid_shapes_var,
                                       fill=st.floats(min_value=-10,
                                                      max_value=10,
                                                      allow_nan=False,
                                                      allow_infinity=False,
                                                      width=32),
                                       elements=st.floats(min_value=-10,
                                                          max_value=10,
                                                          allow_nan=False,
                                                          allow_infinity=False,
                                                          width=32))


class TestColumn2:

    @given(key=st_valid_keys, val=valid_arrays_var_float32)
    @settings(max_examples=200, deadline=None)
    def test_arrayset_variable_shape_float32(self, key, val, variable_shape_repo_co_float32_aset_flat):
        global added_samples

        co = variable_shape_repo_co_float32_aset_flat
        col = co.columns['writtenaset']
        assert col.schema_type == 'variable_shape'

        col[key] = val
        added_samples.add(key)
        out = col[key]
        assert out.dtype == val.dtype
        assert out.shape == val.shape
        assert np.allclose(out, val)
        assert len(col) == len(added_samples)


valid_arrays_var_uint8 = npst.arrays(np.uint8,
                                     shape=valid_shapes_var,
                                     elements=st.integers(min_value=0, max_value=255),
                                     fill=st.integers(min_value=0, max_value=255))


class TestColumn3:

    @given(key=st_valid_keys, val=valid_arrays_var_uint8)
    @settings(max_examples=200, deadline=None)
    def test_arrayset_variable_shape_uint8(self, key, val, variable_shape_repo_co_uint8_aset_flat):
        global added_samples

        co = variable_shape_repo_co_uint8_aset_flat
        col = co.columns['writtenaset']
        assert col.schema_type == 'variable_shape'

        col[key] = val
        added_samples.add(key)
        out = col[key]
        assert out.dtype == val.dtype
        assert out.shape == val.shape
        assert np.allclose(out, val)
        assert len(col) == len(added_samples)


ascii_characters = st.characters(min_codepoint=0, max_codepoint=127)
ascii_text_stratagy = st.text(alphabet=ascii_characters, min_size=0, max_size=500)


class TestColumn4:

    @given(key=st_valid_keys, val=ascii_text_stratagy)
    @settings(max_examples=200, deadline=None)
    def test_str_column_variable_shape(self, key, val, variable_shape_repo_co_str_aset_flat):
        global added_samples

        co = variable_shape_repo_co_str_aset_flat
        col = co.columns['strcolumn']
        col[key] = val

        assert col.schema_type == 'variable_shape'
        assert col.column_type == 'str'
        assert col.dtype == str

        added_samples.add(key)
        out = col[key]
        assert out == val
        assert len(col) == len(added_samples)





import pytest
import numpy as np


def verify_out(capfd, expected):
    out, _ = capfd.readouterr()
    print(out)
    assert expected == out


def test_flat_merge_graph(capfd):
    from hangar.diagnostics import Graph

    flat_log_contents = {
        'head': '3c9530ac0da1106c0acbe1201900c51548bbcdd9',
        'ancestors': {
            '0ff3f2ec156ab8e1026b5271630ccae4556cc260': [''],
            '3c9530ac0da1106c0acbe1201900c51548bbcdd9': ['fed88489ab6e59913aee935169b15fe68755d82c'],
            'fed88489ab6e59913aee935169b15fe68755d82c': ['0ff3f2ec156ab8e1026b5271630ccae4556cc260']},
        'specs': {
            '0ff3f2ec156ab8e1026b5271630ccae4556cc260': {
                'commit_message': 'first commit adding training images and labels',
                'commit_time': 1562203787.257128, 'commit_user': 'Foo User', 'commit_email': 'foo@bar.com'},
            '3c9530ac0da1106c0acbe1201900c51548bbcdd9': {
                'commit_message': 'added testing labels only',
                'commit_time': 1562203787.388417, 'commit_user': 'Foo User', 'commit_email': 'foo@bar.com'},
            'fed88489ab6e59913aee935169b15fe68755d82c': {
                'commit_message': 'added testing images only',
                'commit_time': 1562203787.372292, 'commit_user': 'Foo User', 'commit_email': 'foo@bar.com'}},
        'order': ['3c9530ac0da1106c0acbe1201900c51548bbcdd9',
                'fed88489ab6e59913aee935169b15fe68755d82c',
                '0ff3f2ec156ab8e1026b5271630ccae4556cc260']}

    flat_hash_branch_map = {
        '3c9530ac0da1106c0acbe1201900c51548bbcdd9': ['add-test'],
        '0ff3f2ec156ab8e1026b5271630ccae4556cc260': ['untouched-live-demo-branch']}

    g = Graph(use_color=False)
    g.show_nodes(
        dag=flat_log_contents['ancestors'],
        spec=flat_log_contents['specs'],
        branch=flat_hash_branch_map,
        start=flat_log_contents['head'],
        order=flat_log_contents['order'],
        show_time=False,
        show_user=False)

    expected = '* 3c9530ac0da1106c0acbe1201900c51548bbcdd9 (add-test) : added testing labels only\n'\
               '* fed88489ab6e59913aee935169b15fe68755d82c : added testing images only\n'\
               '* 0ff3f2ec156ab8e1026b5271630ccae4556cc260 (untouched-live-demo-branch) : first commit adding training images and labels\n'

    verify_out(capfd, expected)


def test_three_way_merge_graph(capfd):
    from hangar.diagnostics import Graph

    three_way_log_contents = {
        'head': '074f81d6b9fa5fa856175d47c7cc95cc4a839965',
        'ancestors': {
            '074f81d6b9fa5fa856175d47c7cc95cc4a839965': ['e5ea58dd9c7ffacd45fb128ddc00aced08d13889', '3c9530ac0da1106c0acbe1201900c51548bbcdd9'],
            'e5ea58dd9c7ffacd45fb128ddc00aced08d13889': ['0ff3f2ec156ab8e1026b5271630ccae4556cc260'],
            '0ff3f2ec156ab8e1026b5271630ccae4556cc260': [''],
            '3c9530ac0da1106c0acbe1201900c51548bbcdd9': ['fed88489ab6e59913aee935169b15fe68755d82c'],
            'fed88489ab6e59913aee935169b15fe68755d82c': ['0ff3f2ec156ab8e1026b5271630ccae4556cc260']},
        'specs': {
            '074f81d6b9fa5fa856175d47c7cc95cc4a839965': {
                'commit_message': 'adding in the new testing columns',
                'commit_time': 1562203830.775428, 'commit_user': 'Foo User', 'commit_email': 'foo@bar.com'},
            'e5ea58dd9c7ffacd45fb128ddc00aced08d13889': {
                'commit_message': 'commit adding validation images and labels',
                'commit_time': 1562203787.320624, 'commit_user': 'Foo User', 'commit_email': 'foo@bar.com'},
            '0ff3f2ec156ab8e1026b5271630ccae4556cc260': {
                'commit_message': 'first commit adding training images and labels',
                'commit_time': 1562203787.257128, 'commit_user': 'Foo User', 'commit_email': 'foo@bar.com'},
            '3c9530ac0da1106c0acbe1201900c51548bbcdd9': {
                'commit_message': 'added testing labels only',
                'commit_time': 1562203787.388417, 'commit_user': 'Foo User', 'commit_email': 'foo@bar.com'},
            'fed88489ab6e59913aee935169b15fe68755d82c': {
                'commit_message': 'added testing images only',
                'commit_time': 1562203787.372292, 'commit_user': 'Foo User', 'commit_email': 'foo@bar.com'}},
        'order': ['074f81d6b9fa5fa856175d47c7cc95cc4a839965',
                '3c9530ac0da1106c0acbe1201900c51548bbcdd9',
                'fed88489ab6e59913aee935169b15fe68755d82c',
                'e5ea58dd9c7ffacd45fb128ddc00aced08d13889',
                '0ff3f2ec156ab8e1026b5271630ccae4556cc260']}

    three_way_hash_branch_map = {
        '3c9530ac0da1106c0acbe1201900c51548bbcdd9': ['add-test'],
        'e5ea58dd9c7ffacd45fb128ddc00aced08d13889': ['add-validation'],
        '074f81d6b9fa5fa856175d47c7cc95cc4a839965': ['master'],
        '0ff3f2ec156ab8e1026b5271630ccae4556cc260': ['untouched-live-demo-branch']}

    g = Graph(use_color=False)
    g.show_nodes(
        dag=three_way_log_contents['ancestors'],
        spec=three_way_log_contents['specs'],
        branch=three_way_hash_branch_map,
        start=three_way_log_contents['head'],
        order=three_way_log_contents['order'],
        show_time=False,
        show_user=False)

    real = '*   074f81d6b9fa5fa856175d47c7cc95cc4a839965 (master) : adding in the new testing columns\n'\
           '|\\  \n'\
           '| * 3c9530ac0da1106c0acbe1201900c51548bbcdd9 (add-test) : added testing labels only\n'\
           '| * fed88489ab6e59913aee935169b15fe68755d82c : added testing images only\n'\
           '* | e5ea58dd9c7ffacd45fb128ddc00aced08d13889 (add-validation) : commit adding validation images and labels\n'\
           '|/  \n'\
           '* 0ff3f2ec156ab8e1026b5271630ccae4556cc260 (untouched-live-demo-branch) : first commit adding training images and labels\n'

    verify_out(capfd, real)


def test_octopus_merge_graph(capfd):
    from hangar.diagnostics import Graph

    octopus_log_contents = {
        'head': '05ad17beab54ede8d7f9214c5c6ae44509c3da97',
        'ancestors': {
            '05ad17beab54ede8d7f9214c5c6ae44509c3da97': ['b9c7da873c06c730f52bad5808df5312c4cc0a38', '1b49223ae5e731da3750e4836d14565dbe504f18'],
            'b9c7da873c06c730f52bad5808df5312c4cc0a38': ['a74236e598b96dcde10b176921eb58bb4a9c64bf', 'c4d6875caeff83a29413ae163dbcfdc3c57ad373'],
            'a74236e598b96dcde10b176921eb58bb4a9c64bf': ['9152a4578f74b36838f8187e43c8644b1eba47b5'],
            '9152a4578f74b36838f8187e43c8644b1eba47b5': ['ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d', '21f274d31abc09ede4ad6753f079297885b02a09'],
            'ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d': ['e9ca97e336496b1fceb75869adf0294af5635922'],
            'e9ca97e336496b1fceb75869adf0294af5635922': ['489bceb38246f27cae2a0f47eba0e488d95618db'],
            '489bceb38246f27cae2a0f47eba0e488d95618db': ['17286961175c5cbbd4381fef07cc0a20920a5ce6'],
            '17286961175c5cbbd4381fef07cc0a20920a5ce6': ['63ac654df43bd149a1ca5f919e714bc57e69af99'],
            '63ac654df43bd149a1ca5f919e714bc57e69af99': [''],
            '1b49223ae5e731da3750e4836d14565dbe504f18': ['9152a4578f74b36838f8187e43c8644b1eba47b5'],
            'c4d6875caeff83a29413ae163dbcfdc3c57ad373': ['63ac654df43bd149a1ca5f919e714bc57e69af99'],
            '21f274d31abc09ede4ad6753f079297885b02a09': ['5c0ea20c6513f135f0131d9e10d86801ded29537'],
            '5c0ea20c6513f135f0131d9e10d86801ded29537': ['10e84be056afb2ace6b7ba044ce1e9c9811eae4f'],
            '10e84be056afb2ace6b7ba044ce1e9c9811eae4f': ['e9ca97e336496b1fceb75869adf0294af5635922']},
        'specs': {
            '05ad17beab54ede8d7f9214c5c6ae44509c3da97': {'commit_message': 'try number two',
                'commit_time': 1562363265.6635652, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'b9c7da873c06c730f52bad5808df5312c4cc0a38': {'commit_message': 'merging the long running branch into master',
                'commit_time': 1562363265.652887, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'a74236e598b96dcde10b176921eb58bb4a9c64bf': {'commit_message': 'another on master',
                'commit_time': 1562363265.6346502, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '9152a4578f74b36838f8187e43c8644b1eba47b5': {'commit_message': 'this is the first merge',
                'commit_time': 1562363265.578071, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d': {'commit_message': 'third commit on master',
                'commit_time': 1562363265.4683158, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'e9ca97e336496b1fceb75869adf0294af5635922': {'commit_message': 'second commit on master with training labels',
                'commit_time': 1562363265.398268, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '489bceb38246f27cae2a0f47eba0e488d95618db': {'commit_message': 'second',
                'commit_time': 1562363264.7388191, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '17286961175c5cbbd4381fef07cc0a20920a5ce6': {'commit_message': 'hi',
                'commit_time': 1562363264.735318, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '63ac654df43bd149a1ca5f919e714bc57e69af99': {'commit_message': 'initial commit on master with training images',
                'commit_time': 1562363264.731286, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '1b49223ae5e731da3750e4836d14565dbe504f18': {'commit_message': 'another on try delete',
                'commit_time': 1562363265.642503, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'c4d6875caeff83a29413ae163dbcfdc3c57ad373': {'commit_message': 'first commit on the large branch',
                'commit_time': 1562363265.374819, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '21f274d31abc09ede4ad6753f079297885b02a09': {'commit_message': 'another commit on test banch after adding to new_set',
                'commit_time': 1562363265.56455, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '5c0ea20c6513f135f0131d9e10d86801ded29537': {'commit_message': 'second commit on test branch with new aset',
                'commit_time': 1562363265.545484, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '10e84be056afb2ace6b7ba044ce1e9c9811eae4f': {'commit_message': 'first commit on test branch',
                'commit_time': 1562363265.524131, 'commit_user': 'test user', 'commit_email': 'test@email.com'}},
        'order': [
            '05ad17beab54ede8d7f9214c5c6ae44509c3da97',
            'b9c7da873c06c730f52bad5808df5312c4cc0a38',
            '1b49223ae5e731da3750e4836d14565dbe504f18',
            'a74236e598b96dcde10b176921eb58bb4a9c64bf',
            '9152a4578f74b36838f8187e43c8644b1eba47b5',
            '21f274d31abc09ede4ad6753f079297885b02a09',
            '5c0ea20c6513f135f0131d9e10d86801ded29537',
            '10e84be056afb2ace6b7ba044ce1e9c9811eae4f',
            'ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d',
            'e9ca97e336496b1fceb75869adf0294af5635922',
            'c4d6875caeff83a29413ae163dbcfdc3c57ad373',
            '489bceb38246f27cae2a0f47eba0e488d95618db',
            '17286961175c5cbbd4381fef07cc0a20920a5ce6',
            '63ac654df43bd149a1ca5f919e714bc57e69af99']
        }

    octopus_hash_branch_map = {
        'c4d6875caeff83a29413ae163dbcfdc3c57ad373': ['large_branch'],
        '21f274d31abc09ede4ad6753f079297885b02a09': ['test_branch'],
        '05ad17beab54ede8d7f9214c5c6ae44509c3da97': ['master'],
        '1b49223ae5e731da3750e4836d14565dbe504f18': ['trydelete']}

    g = Graph(use_color=False)
    g.show_nodes(
        dag=octopus_log_contents['ancestors'],
        spec=octopus_log_contents['specs'],
        branch=octopus_hash_branch_map,
        start=octopus_log_contents['head'],
        order=octopus_log_contents['order'],
        show_time=False,
        show_user=False)

    real = '*   05ad17beab54ede8d7f9214c5c6ae44509c3da97 (master) : try number two\n'\
           '|\\  \n'\
           '* \\   b9c7da873c06c730f52bad5808df5312c4cc0a38 : merging the long running branch into master\n'\
           '|\\ \\  \n'\
           '| | * 1b49223ae5e731da3750e4836d14565dbe504f18 (trydelete) : another on try delete\n'\
           '* | | a74236e598b96dcde10b176921eb58bb4a9c64bf : another on master\n'\
           '| |/  \n'\
           '|/|   \n'\
           '* |   9152a4578f74b36838f8187e43c8644b1eba47b5 : this is the first merge\n'\
           '|\\ \\  \n'\
           '| * | 21f274d31abc09ede4ad6753f079297885b02a09 (test_branch) : another commit on test banch after adding to new_set\n'\
           '| * | 5c0ea20c6513f135f0131d9e10d86801ded29537 : second commit on test branch with new aset\n'\
           '| * | 10e84be056afb2ace6b7ba044ce1e9c9811eae4f : first commit on test branch\n'\
           '* | | ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d : third commit on master\n'\
           '|/ /  \n'\
           '* | e9ca97e336496b1fceb75869adf0294af5635922 : second commit on master with training labels\n'\
           '| * c4d6875caeff83a29413ae163dbcfdc3c57ad373 (large_branch) : first commit on the large branch\n'\
           '* | 489bceb38246f27cae2a0f47eba0e488d95618db : second\n'\
           '* | 17286961175c5cbbd4381fef07cc0a20920a5ce6 : hi\n'\
           '|/  \n'\
           '* 63ac654df43bd149a1ca5f919e714bc57e69af99 : initial commit on master with training images\n'

    verify_out(capfd, real)


def test_octopus_large_merge_graph(capfd):
    from hangar.diagnostics import Graph

    octopus_log_contents = {
        'head': 'ddeeff',
        'ancestors': {
            '05ad17beab54ede8d7f9214c5c6ae44509c3da97': ['b9c7da873c06c730f52bad5808df5312c4cc0a38', '1b49223ae5e731da3750e4836d14565dbe504f18'],
            'b9c7da873c06c730f52bad5808df5312c4cc0a38': ['a74236e598b96dcde10b176921eb58bb4a9c64bf', 'c4d6875caeff83a29413ae163dbcfdc3c57ad373', 'e9ca97e336496b1fceb75869adf0294af5635922'],
            'a74236e598b96dcde10b176921eb58bb4a9c64bf': ['9152a4578f74b36838f8187e43c8644b1eba47b5', '21f274d31abc09ede4ad6753f079297885b02a09'],
            '9152a4578f74b36838f8187e43c8644b1eba47b5': ['ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d', '21f274d31abc09ede4ad6753f079297885b02a09'],
            'ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d': ['e9ca97e336496b1fceb75869adf0294af5635922', 'c4d6875caeff83a29413ae163dbcfdc3c57ad373'],
            'e9ca97e336496b1fceb75869adf0294af5635922': ['489bceb38246f27cae2a0f47eba0e488d95618db'],
            '489bceb38246f27cae2a0f47eba0e488d95618db': ['17286961175c5cbbd4381fef07cc0a20920a5ce6'],
            '17286961175c5cbbd4381fef07cc0a20920a5ce6': ['63ac654df43bd149a1ca5f919e714bc57e69af99'],
            '63ac654df43bd149a1ca5f919e714bc57e69af99': [''],
            '1b49223ae5e731da3750e4836d14565dbe504f18': ['9152a4578f74b36838f8187e43c8644b1eba47b5', 'a74236e598b96dcde10b176921eb58bb4a9c64bf'],
            'c4d6875caeff83a29413ae163dbcfdc3c57ad373': ['63ac654df43bd149a1ca5f919e714bc57e69af99'],
            '21f274d31abc09ede4ad6753f079297885b02a09': ['5c0ea20c6513f135f0131d9e10d86801ded29537'],
            '5c0ea20c6513f135f0131d9e10d86801ded29537': ['10e84be056afb2ace6b7ba044ce1e9c9811eae4f', 'ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d'],
            '10e84be056afb2ace6b7ba044ce1e9c9811eae4f': ['e9ca97e336496b1fceb75869adf0294af5635922', 'c4d6875caeff83a29413ae163dbcfdc3c57ad373'],
            'aabbcc': ['9152a4578f74b36838f8187e43c8644b1eba47b5', '5c0ea20c6513f135f0131d9e10d86801ded29537'],
            'ddeeff': ['aabbcc', '05ad17beab54ede8d7f9214c5c6ae44509c3da97'],
        },
        'specs': {
            'ddeeff': {'commit_message': 'new master',
                'commit_time': 1562363266.6635652, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '05ad17beab54ede8d7f9214c5c6ae44509c3da97': {'commit_message': 'try number two',
                'commit_time': 1562363265.6635652, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'b9c7da873c06c730f52bad5808df5312c4cc0a38': {'commit_message': 'merging the long running branch into master',
                'commit_time': 1562363265.652887, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'a74236e598b96dcde10b176921eb58bb4a9c64bf': {'commit_message': 'another on master',
                'commit_time': 1562363265.6346502, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '9152a4578f74b36838f8187e43c8644b1eba47b5': {'commit_message': 'this is the first merge',
                'commit_time': 1562363265.578071, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d': {'commit_message': 'third commit on master',
                'commit_time': 1562363265.4683158, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'e9ca97e336496b1fceb75869adf0294af5635922': {'commit_message': 'second commit on master with training labels',
                'commit_time': 1562363265.398268, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '489bceb38246f27cae2a0f47eba0e488d95618db': {'commit_message': 'second',
                'commit_time': 1562363264.7388191, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '17286961175c5cbbd4381fef07cc0a20920a5ce6': {'commit_message': 'hi',
                'commit_time': 1562363264.735318, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '63ac654df43bd149a1ca5f919e714bc57e69af99': {'commit_message': 'initial commit on master with training images',
                'commit_time': 1562363264.731286, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '1b49223ae5e731da3750e4836d14565dbe504f18': {'commit_message': 'another on try delete',
                'commit_time': 1562363265.642503, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'aabbcc': {'commit_message': 'made up b',
                'commit_time': 1562363265.640021, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            'c4d6875caeff83a29413ae163dbcfdc3c57ad373': {'commit_message': 'first commit on the large branch',
                'commit_time': 1562363265.374819, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '21f274d31abc09ede4ad6753f079297885b02a09': {'commit_message': 'another commit on test banch after adding to new_set',
                'commit_time': 1562363265.56455, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '5c0ea20c6513f135f0131d9e10d86801ded29537': {'commit_message': 'second commit on test branch with new aset',
                'commit_time': 1562363265.545484, 'commit_user': 'test user', 'commit_email': 'test@email.com'},
            '10e84be056afb2ace6b7ba044ce1e9c9811eae4f': {'commit_message': 'first commit on test branch',
                'commit_time': 1562363265.524131, 'commit_user': 'test user', 'commit_email': 'test@email.com'}},
        'order': [
            'ddeeff',
            '05ad17beab54ede8d7f9214c5c6ae44509c3da97',
            'b9c7da873c06c730f52bad5808df5312c4cc0a38',
            '1b49223ae5e731da3750e4836d14565dbe504f18',
            'aabbcc',
            'a74236e598b96dcde10b176921eb58bb4a9c64bf',
            '9152a4578f74b36838f8187e43c8644b1eba47b5',
            '21f274d31abc09ede4ad6753f079297885b02a09',
            '5c0ea20c6513f135f0131d9e10d86801ded29537',
            '10e84be056afb2ace6b7ba044ce1e9c9811eae4f',
            'ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d',
            'e9ca97e336496b1fceb75869adf0294af5635922',
            'c4d6875caeff83a29413ae163dbcfdc3c57ad373',
            '489bceb38246f27cae2a0f47eba0e488d95618db',
            '17286961175c5cbbd4381fef07cc0a20920a5ce6',
            '63ac654df43bd149a1ca5f919e714bc57e69af99']
        }

    octopus_hash_branch_map = {
        'c4d6875caeff83a29413ae163dbcfdc3c57ad373': ['large_branch'],
        '21f274d31abc09ede4ad6753f079297885b02a09': ['test_branch'],
        'ddeeff': ['master'],
        '1b49223ae5e731da3750e4836d14565dbe504f18': ['trydelete'],
        'aabbcc': ['madeupbranch']
    }

    g = Graph(use_color=False)
    g.show_nodes(
        dag=octopus_log_contents['ancestors'],
        spec=octopus_log_contents['specs'],
        branch=octopus_hash_branch_map,
        start=octopus_log_contents['head'],
        order=octopus_log_contents['order'],
        show_time=True,
        show_user=True)

    real = '*   ddeeff (master) (05Jul2019 21:47:46)(test user): new master\n'\
           '|\\  \n'\
           '| *   05ad17beab54ede8d7f9214c5c6ae44509c3da97 (05Jul2019 21:47:45)(test user): try number two\n'\
           '| |\\  \n'\
           '| | \\     \n'\
           '| |  \\    \n'\
           '| *-. \\   b9c7da873c06c730f52bad5808df5312c4cc0a38 (05Jul2019 21:47:45)(test user): merging the long running branch into master\n'\
           '| |\\ \\ \\  \n'\
           '| | | | *   1b49223ae5e731da3750e4836d14565dbe504f18 (trydelete) (05Jul2019 21:47:45)(test user): another on try delete\n'\
           '| | | | |\\  \n'\
           '| | |_|_|/  \n'\
           '| |/| | |   \n'\
           '* | | | |   aabbcc (madeupbranch) (05Jul2019 21:47:45)(test user): made up b\n'\
           '|\\ \\ \\ \\ \\  \n'\
           '| |_|_|_|/  \n'\
           '|/| | | |   \n'\
           '| | * | |   a74236e598b96dcde10b176921eb58bb4a9c64bf (05Jul2019 21:47:45)(test user): another on master\n'\
           '| | |\\ \\ \\  \n'\
           '| |/ / / /  \n'\
           '|/| | | |   \n'\
           '* | | | |   9152a4578f74b36838f8187e43c8644b1eba47b5 (05Jul2019 21:47:45)(test user): this is the first merge\n'\
           '|\\ \\ \\ \\ \\  \n'\
           '| | |/ / /  \n'\
           '| |/| | |   \n'\
           '| * | | | 21f274d31abc09ede4ad6753f079297885b02a09 (test_branch) (05Jul2019 21:47:45)(test user): another commit on test banch after adding to new_set\n'\
           '| |/ / /  \n'\
           '| * | |   5c0ea20c6513f135f0131d9e10d86801ded29537 (05Jul2019 21:47:45)(test user): second commit on test branch with new aset\n'\
           '| |\\ \\ \\  \n'\
           '| |/ / /  \n'\
           '|/| | |   \n'\
           '| * | |   10e84be056afb2ace6b7ba044ce1e9c9811eae4f (05Jul2019 21:47:45)(test user): first commit on test branch\n'\
           '| |\\ \\ \\  \n'\
           '| | |/ /  \n'\
           '| | | /   \n'\
           '| | |/    \n'\
           '| |/|     \n'\
           '* | |   ef7b6e5bcaaebf62b9e02902ff60eb7862c3472d (05Jul2019 21:47:45)(test user): third commit on master\n'\
           '|\\ \\ \\  \n'\
           '| |/ /  \n'\
           '|/| /   \n'\
           '| |/    \n'\
           '* | e9ca97e336496b1fceb75869adf0294af5635922 (05Jul2019 21:47:45)(test user): second commit on master with training labels\n'\
           '| * c4d6875caeff83a29413ae163dbcfdc3c57ad373 (large_branch) (05Jul2019 21:47:45)(test user): first commit on the large branch\n'\
           '* | 489bceb38246f27cae2a0f47eba0e488d95618db (05Jul2019 21:47:44)(test user): second\n'\
           '* | 17286961175c5cbbd4381fef07cc0a20920a5ce6 (05Jul2019 21:47:44)(test user): hi\n'\
           '|/  \n'\
           '* 63ac654df43bd149a1ca5f919e714bc57e69af99 (05Jul2019 21:47:44)(test user): initial commit on master with training images\n'

    verify_out(capfd, real)


def test_repo_log_return_contents_correct_default_args(repo):

    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co['test_meta']['foo'] = 'bar'
    ancestor_digest = co.commit('first')
    co['test_meta']['hello'] = 'world'
    master_head = co.commit('second')
    co.close()

    ancestor_branch = repo.create_branch('ancestor', base_commit=ancestor_digest)
    dev_branch = repo.create_branch('dev', base_commit=ancestor_digest)

    co = repo.checkout(write=True, branch=dev_branch.name)
    co['test_meta']['zen'] = 'of python'
    dev_head = co.commit('third on test')
    co.close()

    log = repo.log(return_contents=True)

    assert log['head'] == dev_head

    expected_ancestors = {
        dev_head: [ancestor_digest],
        ancestor_digest: [''],
    }
    assert log['ancestors'] == expected_ancestors

    assert len(log['specs']) == 2
    assert len(log['specs'][ancestor_digest]) == 4
    assert len(log['specs'][dev_head]) == 4
    assert log['specs'][ancestor_digest]['commit_message'] == 'first'
    assert log['specs'][dev_head]['commit_message'] == 'third on test'

    assert log['order'] == [dev_head, ancestor_digest]

    assert len(log['branch_heads']) == 2
    assert log['branch_heads'][ancestor_digest] == [ancestor_branch.name]
    assert log['branch_heads'][dev_head] == [dev_branch.name]


def test_repo_log_return_contents_correct_when_specify_branch_name(repo):

    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co['test_meta']['foo'] = 'bar'
    ancestor_digest = co.commit('first')
    co['test_meta']['hello'] = 'world'
    master_head = co.commit('second')
    co.close()

    ancestor_branch = repo.create_branch('ancestor', base_commit=ancestor_digest)
    dev_branch = repo.create_branch('dev', base_commit=ancestor_digest)

    co = repo.checkout(write=True, branch=dev_branch.name)
    co['test_meta']['zen'] = 'of python'
    dev_head = co.commit('third on test')
    co.close()

    log = repo.log(branch='master', return_contents=True)

    assert log['head'] == master_head

    expected_ancestors = {
        master_head: [ancestor_digest],
        ancestor_digest: [''],
    }
    assert log['ancestors'] == expected_ancestors

    assert len(log['specs']) == 2
    assert len(log['specs'][ancestor_digest]) == 4
    assert len(log['specs'][master_head]) == 4
    assert log['specs'][ancestor_digest]['commit_message'] == 'first'
    assert log['specs'][master_head]['commit_message'] == 'second'

    assert log['order'] == [master_head, ancestor_digest]

    assert len(log['branch_heads']) == 2
    assert log['branch_heads'][ancestor_digest] == [ancestor_branch.name]
    assert log['branch_heads'][master_head] == ['master']


def test_repo_log_return_contents_correct_when_specify_digest(repo):

    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    co['test_meta']['foo'] = 'bar'
    ancestor_digest = co.commit('first')
    co['test_meta']['hello'] = 'world'
    master_head = co.commit('second')
    co.close()

    ancestor_branch = repo.create_branch('ancestor', base_commit=ancestor_digest)
    dev_branch = repo.create_branch('dev', base_commit=ancestor_digest)

    co = repo.checkout(write=True, branch=dev_branch.name)
    co['test_meta']['zen'] = 'of python'
    dev_head = co.commit('third on test')
    co.close()

    log = repo.log(commit=master_head, return_contents=True)

    assert log['head'] == master_head

    expected_ancestors = {
        master_head: [ancestor_digest],
        ancestor_digest: [''],
    }
    assert log['ancestors'] == expected_ancestors

    assert len(log['specs']) == 2
    assert len(log['specs'][ancestor_digest]) == 4
    assert len(log['specs'][master_head]) == 4
    assert log['specs'][ancestor_digest]['commit_message'] == 'first'
    assert log['specs'][master_head]['commit_message'] == 'second'

    assert log['order'] == [master_head, ancestor_digest]

    assert len(log['branch_heads']) == 2
    assert log['branch_heads'][ancestor_digest] == [ancestor_branch.name]
    assert log['branch_heads'][master_head] == ['master']



import atexit
import numpy as np
import pytest
from conftest import fixed_shape_backend_params


class TestCheckout(object):

    def test_write_checkout_specifying_commit_not_allowed_if_commit_exists(self, aset_samples_initialized_repo):
        cmt_digest = aset_samples_initialized_repo.log(return_contents=True)['head']
        with pytest.raises(ValueError):
            aset_samples_initialized_repo.checkout(write=True, commit=cmt_digest)

    def test_write_checkout_specifying_commit_not_allowed_if_commit_does_not_exists(self, aset_samples_initialized_repo):
        cmt_digest = 'notrealcommit'
        with pytest.raises(ValueError):
            aset_samples_initialized_repo.checkout(write=True, commit=cmt_digest)

    def test_two_write_checkouts(self, repo):
        w1_checkout = repo.checkout(write=True)
        with pytest.raises(PermissionError):
            repo.checkout(write=True)
        w1_checkout.close()

    def test_two_read_checkouts(self, repo, array5by7):
        w_checkout = repo.checkout(write=True)
        arrayset_name = 'aset'
        r_ds = w_checkout.add_ndarray_column(name=arrayset_name, prototype=array5by7)
        r_ds['1'] = array5by7
        w_checkout.commit('init')
        r1_checkout = repo.checkout()
        r2_checkout = repo.checkout()
        assert np.allclose(r1_checkout.columns['aset']['1'], array5by7)
        assert np.allclose(
            r1_checkout.columns['aset']['1'], r2_checkout.columns['aset']['1'])
        r1_checkout.close()
        r2_checkout.close()
        w_checkout.close()

    def test_write_with_read_checkout(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout()
        with pytest.raises(AttributeError):
            co.add_ndarray_column(name='aset', shape=(5, 7), dtype=np.float64)
        with pytest.raises(AttributeError):
            co.add_str_column('test_meta')
        co.close()

    def test_writer_aset_obj_not_accessible_after_close(self, two_commit_filled_samples_repo):
        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=True)
        asets = co.columns
        aset = co.columns['writtenaset']
        co.close()

        with pytest.raises(PermissionError):
            asets.iswriteable
        with pytest.raises(PermissionError):
            shouldFail = asets['writtenaset']
        with pytest.raises(PermissionError):
            aset.iswriteable

    def test_writer_aset_obj_arrayset_iter_values_not_accessible_after_close(self, two_commit_filled_samples_repo):
        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=True)
        oldObjs = []
        for oldObj in co.columns.values():
            oldObjs.append(oldObj)
        co.close()

        for oldObj in oldObjs:
            with pytest.raises(PermissionError):
                oldObj.column

    def test_writer_aset_obj_arrayset_iter_items_not_accessible_after_close(self, two_commit_filled_samples_repo):
        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=True)
        oldObjs = {}
        for oldName, oldObj in co.columns.items():
            oldObjs[oldName] = oldObj
        co.close()

        for name, obj in oldObjs.items():
            assert isinstance(name, str)
            with pytest.raises(PermissionError):
                obj.column

    def test_writer_aset_obj_not_accessible_after_commit_and_close(self, aset_samples_initialized_repo, array5by7):
        repo = aset_samples_initialized_repo
        co = repo.checkout(write=True)
        asets = co.columns
        aset = co.columns['writtenaset']
        aset['1'] = array5by7
        co.commit('hey there')
        co.close()

        with pytest.raises(PermissionError):
            asets.iswriteable
        with pytest.raises(PermissionError):
            shouldFail = asets['writtenaset']
        with pytest.raises(PermissionError):
            aset.iswriteable
        with pytest.raises(PermissionError):
            shouldFail = aset['1']

    def test_reader_aset_obj_not_accessible_after_close(self, two_commit_filled_samples_repo):
        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=False)
        asets = co.columns
        aset = co.columns['writtenaset']
        co.close()

        with pytest.raises(PermissionError):
            asets.iswriteable
        with pytest.raises(PermissionError):
            shouldFail = asets['writtenaset']
        with pytest.raises(PermissionError):
            aset.iswriteable

    def test_reader_aset_obj_column_iter_values_not_accessible_after_close(self, two_commit_filled_samples_repo):
        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=False)
        oldObjs = []
        for oldObj in co.columns.values():
            oldObjs.append(oldObj)
        co.close()

        for oldObj in oldObjs:
            with pytest.raises(PermissionError):
                oldObj.column

    def test_reader_aset_obj_arrayset_iter_items_not_accessible_after_close(self, two_commit_filled_samples_repo):
        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=False)
        oldObjs = {}
        for oldName, oldObj in co.columns.items():
            oldObjs[oldName] = oldObj
        co.close()

        for name, obj in oldObjs.items():
            assert isinstance(name, str)
            with pytest.raises(PermissionError):
                obj.column

    def test_reader_arrayset_context_manager_not_accessible_after_close(self, two_commit_filled_samples_repo):
        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=False)
        aset = co.columns['writtenaset']
        klist = []
        with aset as ds:
            for k in ds.keys():
                klist.append(k)
                a = ds
        co.close()

        with pytest.raises(PermissionError):
            a.column
        with pytest.raises(PermissionError):
            ds.column
        with pytest.raises(PermissionError):
            aset[klist[0]]

    def test_writer_arrayset_context_manager_not_accessible_after_close(self, two_commit_filled_samples_repo):
        repo = two_commit_filled_samples_repo
        co = repo.checkout(write=True)
        aset = co.columns['writtenaset']
        with aset as ds:
            # for k in ds.keys():
            #     klist.append(k)
            a = ds
            a['1232'] = np.random.randn(5, 7).astype(np.float32)
        co.close()

        with pytest.raises(PermissionError):
            a.column
        with pytest.raises(PermissionError):
            ds.column
        with pytest.raises(PermissionError):
            aset['1232']

    def test_close_read_does_not_invalidate_write_checkout(self, aset_samples_initialized_repo, array5by7):
        repo = aset_samples_initialized_repo
        r_co = repo.checkout(write=False)
        w_co = repo.checkout(write=True)
        r_co.close()

        with pytest.raises(PermissionError):
            shouldFail = r_co.columns

        aset = w_co.columns['writtenaset']
        aset['1'] = array5by7
        assert np.allclose(w_co.columns['writtenaset']['1'], array5by7)
        w_co.commit('hello commit')
        w_co.close()

        with pytest.raises(PermissionError):
            aset.column

    def test_close_write_does_not_invalidate_read_checkout(self, aset_samples_initialized_repo, array5by7):
        repo = aset_samples_initialized_repo
        r_co = repo.checkout(write=False)
        w_co = repo.checkout(write=True)

        aset = w_co.columns['writtenaset']
        aset['1'] = array5by7
        assert np.allclose(w_co.columns['writtenaset']['1'], array5by7)
        w_co.commit('hello commit')
        w_co.close()

        assert 'writtenaset' in r_co.columns
        with pytest.raises(PermissionError):
            aset.column
        r_co.close()
        with pytest.raises(PermissionError):
            r_co.columns

    def test_operate_on_arrayset_after_closing_old_checkout(self, repo, array5by7):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', prototype=array5by7)
        co.commit('this is a commit message')
        co.close()
        co = repo.checkout(write=True)
        with pytest.raises(PermissionError):
            aset['1'] = array5by7
            co.commit('this is a commit message')
        co.close()
        with pytest.raises(PermissionError):
            aset['1']

    def test_operate_on_closed_checkout(self, repo, array5by7):
        co = repo.checkout(write=True)
        co.add_ndarray_column('aset', prototype=array5by7)
        co.commit('this is a commit message')
        co.close()
        with pytest.raises(PermissionError):
            co.columns['aset']['1'] = array5by7

    @pytest.mark.parametrize("aset_backend", fixed_shape_backend_params)
    def test_operate_on_arrayset_samples_after_commiting_but_not_closing_checkout(self, aset_backend, repo, array5by7):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', prototype=array5by7, backend=aset_backend)
        aset['1'] = array5by7
        co.commit('hi')

        aset['2'] = array5by7  # this raises Exception since the reference to aset i gon
        co.commit('hello 2')
        assert np.allclose(aset['2'], array5by7)
        co.close()

        with pytest.raises(PermissionError):
            aset.name

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    def test_operate_on_arraysets_after_commiting_but_not_closing_checkout(self, aset1_backend, aset2_backend, repo, array5by7):
        co = repo.checkout(write=True)
        asets = co.columns
        aset = co.add_ndarray_column('aset', prototype=array5by7, backend=aset1_backend)
        aset['1'] = array5by7
        co.commit('hi')

        aset2 = co.add_ndarray_column('arange', prototype=np.arange(50), backend=aset2_backend)
        aset2['0'] = np.arange(50)
        co.commit('hello 2')
        assert np.allclose(aset2['0'], np.arange(50))
        co.close()

        with pytest.raises(PermissionError):
            co.columns
        with pytest.raises(PermissionError):
            asets.iswriteable
        with pytest.raises(PermissionError):
            aset2.name

    def test_with_wrong_argument_value(self, repo):
        # It is intuitive to a user to pass branchname as positional
        # argument but hangar expect permission as first argument
        with pytest.raises(ValueError):
            repo.checkout('branchname')
        with pytest.raises(ValueError):
            repo.checkout(write='True')
        with pytest.raises(ValueError):
            repo.checkout(branch=True)
        co = repo.checkout(True)  # This should not raise any excpetion
        # unregister close operation as conftest will close env before this is called.
        atexit.unregister(co.close)


    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    def test_reset_staging_area_no_changes_made_does_not_work(self, aset1_backend, aset2_backend, repo, array5by7):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', prototype=array5by7, backend=aset1_backend)
        aset2 = co.add_ndarray_column('arange', prototype=np.arange(50), backend=aset2_backend)
        aset['1'] = array5by7
        aset2['0'] = np.arange(50)
        co.commit('hi')

        # verifications before reset
        assert np.allclose(aset2['0'], np.arange(50))
        assert len(co.columns) == 2
        assert co.columns['arange'].iswriteable

        with pytest.raises(RuntimeError, match='No changes made'):
            co.reset_staging_area()

        # verifications after reset
        assert np.allclose(aset2['0'], np.arange(50))
        assert len(co.columns) == 2
        assert co.columns['arange'].iswriteable
        co.close()

    @pytest.mark.parametrize("aset1_backend", fixed_shape_backend_params)
    @pytest.mark.parametrize("aset2_backend", fixed_shape_backend_params)
    def test_reset_staging_area_clears_arraysets(self, aset1_backend, aset2_backend, repo, array5by7):
        co = repo.checkout(write=True)
        aset = co.add_ndarray_column('aset', prototype=array5by7, backend=aset1_backend)
        aset['1'] = array5by7
        co.commit('hi')

        aset2 = co.add_ndarray_column('arange', prototype=np.arange(50), backend=aset2_backend)
        aset2['0'] = np.arange(50)
        # verifications before reset
        assert np.allclose(aset2['0'], np.arange(50))
        assert len(co.columns) == 2
        assert co.columns['arange'].iswriteable

        co.reset_staging_area()
        # behavior expected after reset
        assert len(co.columns) == 1
        with pytest.raises(PermissionError):
            aset2['0']
        with pytest.raises(KeyError):
            co.columns['arange']
        co.close()

    @pytest.mark.parametrize('write', [True, False])
    def test_checkout_dunder_contains_method(self, repo_20_filled_samples, write):
        co = repo_20_filled_samples.checkout(write=write)
        assert 'writtenaset' in co
        assert 'second_aset' in co
        assert 'doesnotexist' not in co
        co.close()

    @pytest.mark.parametrize('write', [True, False])
    def test_checkout_dunder_len_method(self, repo_20_filled_samples, write):
        co = repo_20_filled_samples.checkout(write=write)
        assert len(co) == 2
        co.close()

    @pytest.mark.parametrize('write', [True, False])
    def test_checkout_dunder_iter_method(self, repo_20_filled_samples, write):
        from typing import Iterable
        co = repo_20_filled_samples.checkout(write=write)
        it = iter(co)
        assert isinstance(it, Iterable)
        icount = 0
        for k in it:
            assert k in ['writtenaset', 'second_aset']
            icount += 1
        assert icount == 2
        co.close()

    @pytest.mark.parametrize('write', [True, False])
    def test_checkout_keys_method(self, repo_20_filled_samples, write):
        co = repo_20_filled_samples.checkout(write=write)
        keys = list(co.keys())
        assert len(keys) == 2
        for k in ['writtenaset', 'second_aset']:
            assert k in keys
        co.close()

    @pytest.mark.parametrize('write', [True, False])
    def test_checkout_values_method(self, repo_20_filled_samples, write):
        from hangar.columns.layout_nested import NestedSampleWriter, NestedSampleReader
        from hangar.columns.layout_flat import FlatSampleWriter, FlatSampleReader
        possible_classes = (
            NestedSampleWriter, NestedSampleReader, FlatSampleReader, FlatSampleWriter)

        co = repo_20_filled_samples.checkout(write=write)
        icount = 0
        for col in co.values():
            assert isinstance(col, possible_classes)
            icount += 1
        assert icount == 2
        co.close()

    @pytest.mark.parametrize('write', [True, False])
    def test_checkout_items_method(self, repo_20_filled_samples, write):
        from hangar.columns.layout_nested import NestedSampleWriter, NestedSampleReader
        from hangar.columns.layout_flat import FlatSampleWriter, FlatSampleReader
        possible_classes = (
            NestedSampleWriter, NestedSampleReader, FlatSampleReader, FlatSampleWriter)

        co = repo_20_filled_samples.checkout(write=write)
        icount = 0
        for k, col in co.items():
            assert k in ['writtenaset', 'second_aset']
            assert isinstance(col, possible_classes)
            icount += 1
        assert icount == 2
        co.close()

    @pytest.mark.parametrize('write', [True, False])
    def test_checkout_log_method(self, repo_20_filled_samples, write):
        repo_log = repo_20_filled_samples.log(return_contents=True)
        co = repo_20_filled_samples.checkout(write=write)
        co_log = co.log(return_contents=True)
        co.close()
        assert repo_log == co_log


class TestBranchingMergingInCheckout(object):

    def test_merge(self, aset_samples_initialized_repo, array5by7):
        branch = aset_samples_initialized_repo.create_branch('testbranch')
        assert isinstance(branch.name, str)
        assert isinstance(branch.digest, str)
        co = aset_samples_initialized_repo.checkout(write=True, branch=branch.name)
        assert co._branch_name == branch.name
        co.add_str_column('test_meta')
        co.columns['writtenaset']['1'] = array5by7
        co['test_meta'].update({'a': 'b'})
        co.commit('this is a commit message')
        co.close()
        aset_samples_initialized_repo.merge('test merge', 'master', branch.name)
        co = aset_samples_initialized_repo.checkout()
        assert (co.columns['writtenaset']['1'] == array5by7).all()
        assert co['test_meta'].get('a') == 'b'
        co.close()

    def test_merge_without_closing_previous_checkout(self, aset_samples_initialized_repo, array5by7):
        branch = aset_samples_initialized_repo.create_branch('testbranch')
        co = aset_samples_initialized_repo.checkout(write=True, branch=branch.name)
        co.columns['writtenaset']['1'] = array5by7
        co.commit('this is a commit message')
        with pytest.raises(PermissionError):
            aset_samples_initialized_repo.merge('test merge', 'master', branch.name)
        # unregister close operation as conftest will close env before this is called.
        atexit.unregister(co.close)

    def test_merge_multiple_checkouts_same_aset(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.add_str_column('test_meta')
        co.commit('test meta commit')
        co.close()
        branch1 = aset_samples_initialized_repo.create_branch('testbranch1')
        co = aset_samples_initialized_repo.checkout(write=True, branch=branch1.name)
        co.columns['writtenaset']['1'] = array5by7
        co['test_meta'].update({'a1': 'b1'})
        co.commit('this is a commit message')
        co.close()

        branch2 = aset_samples_initialized_repo.create_branch('testbranch2')
        co = aset_samples_initialized_repo.checkout(write=True, branch=branch2.name)
        co.columns['writtenaset']['2'] = array5by7
        co['test_meta'].update({'a2': 'b2'})
        co.commit('this is a commit message')
        co.close()

        aset_samples_initialized_repo.merge('test merge 1', 'master', branch1.name)
        aset_samples_initialized_repo.merge('test merge 2', 'master', branch2.name)

        co = aset_samples_initialized_repo.checkout(branch='master')
        assert len(co.columns) == 2
        assert len(co.columns['writtenaset']) == 2
        assert list(co['test_meta'].keys()) == ['a1', 'a2']
        co.close()

    def test_merge_multiple_checkouts_multiple_aset(self, aset_samples_initialized_repo, array5by7):
        branch1 = aset_samples_initialized_repo.create_branch('testbranch1')
        co = aset_samples_initialized_repo.checkout(write=True, branch=branch1.name)
        co.columns['writtenaset']['1'] = array5by7
        co.commit('this is a commit message')
        co.close()

        branch2 = aset_samples_initialized_repo.create_branch('testbranch2')
        co = aset_samples_initialized_repo.checkout(write=True, branch=branch2.name)
        second_aset = co.add_ndarray_column(name='second_aset', prototype=array5by7)
        second_aset['1'] = array5by7
        co.commit('this is a commit message')
        co.close()

        aset_samples_initialized_repo.merge('test merge 1', 'master', branch1.name)
        aset_samples_initialized_repo.merge('test merge 2', 'master', branch2.name)

        co = aset_samples_initialized_repo.checkout(branch='master')
        assert len(co.columns) == 2
        assert len(co.columns['writtenaset']) == 1
        assert len(co.columns['second_aset']) == 1
        co.close()

    def test_merge_diverged_conflict(self, aset_samples_initialized_repo, array5by7):
        co = aset_samples_initialized_repo.checkout(write=True)
        co.add_str_column('test_meta')
        co.commit('test meta commit')
        co.close()
        branch1 = aset_samples_initialized_repo.create_branch('testbranch1')
        branch2 = aset_samples_initialized_repo.create_branch('testbranch2')

        co = aset_samples_initialized_repo.checkout(write=True, branch=branch1.name)
        co.columns['writtenaset']['1'] = array5by7
        co['test_meta'].update({'a': 'b'})
        co.commit('this is a commit message')
        co.close()

        co = aset_samples_initialized_repo.checkout(write=True, branch=branch2.name)
        newarray = np.zeros_like(array5by7)
        co.columns['writtenaset']['1'] = newarray
        co['test_meta'].update({'a': 'c'})
        co.commit('this is a commit message')
        co.close()

        aset_samples_initialized_repo.merge('commit message', 'master', branch1.name)

        with pytest.raises(ValueError):
            aset_samples_initialized_repo.merge('commit message', 'master', branch2.name)

    def test_new_branch_from_where(self, aset_samples_initialized_repo, array5by7):
        branch1 = aset_samples_initialized_repo.create_branch('testbranch1')
        branch2 = aset_samples_initialized_repo.create_branch('testbranch2')
        co1 = aset_samples_initialized_repo.checkout(write=True, branch=branch1.name)
        h1 = aset_samples_initialized_repo.log(branch=co1.branch_name, return_contents=True)
        co1.close()

        co2 = aset_samples_initialized_repo.checkout(write=True, branch=branch2.name)
        co2.add_ndarray_column('aset2', prototype=array5by7)
        co2.columns['aset2']['2'] = array5by7
        co2.commit('this is a merge message')
        co2.close()
        h2 = aset_samples_initialized_repo.log(branch=branch2.name, return_contents=True)

        branch3 = aset_samples_initialized_repo.create_branch('testbranch3')
        co3 = aset_samples_initialized_repo.checkout(write=True, branch=branch3.name)
        h3 = aset_samples_initialized_repo.log(branch=co3.branch_name, return_contents=True)
        co3.close()

        assert h2['head'] == h3['head']
        assert h2['ancestors'][h2['head']] == h3['ancestors'][h3['head']]
        assert h1['head'] in h2['ancestors'][h2['head']]

    def test_cannot_checkout_branch_with_staged_changes(self, aset_samples_initialized_repo, array5by7):
        branch1 = aset_samples_initialized_repo.create_branch('testbranch1')
        branch2 = aset_samples_initialized_repo.create_branch('testbranch2')
        co1 = aset_samples_initialized_repo.checkout(write=True, branch=branch1.name)
        initial_cmt = co1.commit_hash
        co1.add_ndarray_column('aset2', prototype=array5by7)
        co1.columns['aset2']['2'] = array5by7
        co1.close()

        with pytest.raises(ValueError):
            con = aset_samples_initialized_repo.checkout(write=True, branch=branch2.name)

        co1 = aset_samples_initialized_repo.checkout(write=True, branch=branch1.name)
        co1.commit('hi')
        assert co1.commit_hash != initial_cmt
        assert co1.branch_name == branch1.name
        co1.close()

        co2 = aset_samples_initialized_repo.checkout(write=True, branch=branch2.name)
        assert co2.branch_name == branch2.name
        assert co2.commit_hash == initial_cmt
        co2.close()


def test_full_from_short_commit_digest(two_commit_filled_samples_repo):
    from hangar.records.commiting import expand_short_commit_digest

    repo = two_commit_filled_samples_repo
    history = repo.log(branch='master', return_contents=True)
    commits = history['order']
    for full_cmt in commits:
        short_cmt = full_cmt[:18]
        found_cmt = expand_short_commit_digest(repo._env.refenv, short_cmt)
        assert found_cmt == full_cmt

    with pytest.raises(KeyError, match='No matching commit hash found starting with'):
        expand_short_commit_digest(repo._env.refenv, 'zzzzzzzzzzzzzzzzzzzzzzzzzzzz')


def test_writer_context_manager_objects_are_gc_removed_after_co_close(two_commit_filled_samples_repo):

    repo = two_commit_filled_samples_repo
    co = repo.checkout(write=True)
    co.add_str_column('test_meta')
    with co['test_meta'] as m:
        m['aa'] = 'bb'
        cmt1 = co.commit('here is the first commit')
        with co.columns['writtenaset'] as d:
            d['2422'] = d['0'] + 213
            cmt2 = co.commit('here is the second commit')

    assert co.close() is None
    with pytest.raises(PermissionError):
        _ = m.__dict__
    with pytest.raises(PermissionError):
        _ = d.column
    with pytest.raises(PermissionError):
        _ = co.columns
    assert co.__dict__ == {}

    co = repo.checkout(commit=cmt1)
    assert 'aa' in co['test_meta']
    assert co['test_meta']['aa'] == 'bb'
    co.close()

    co = repo.checkout(commit=cmt2)
    assert 'aa' in co['test_meta']
    assert co['test_meta']['aa'] == 'bb'
    assert '2422' in co.columns['writtenaset']
    assert np.allclose(co.columns['writtenaset']['2422'],
                       co.columns['writtenaset']['0'] + 213)
    co.close()


def test_reader_context_manager_objects_are_gc_removed_after_co_close(two_commit_filled_samples_repo):

    repo = two_commit_filled_samples_repo
    co = repo.checkout(write=False)
    with co.columns['writtenaset'] as d:
        ds = d['2']

    assert d.iswriteable is False
    assert np.allclose(ds, d.get('2'))
    assert np.allclose(ds, co.columns['writtenaset'].get('2'))

    assert co.close() is None

    with pytest.raises(PermissionError):
        d.column
    with pytest.raises(AttributeError):
        co._columns
    with pytest.raises(PermissionError):
        str(co.columns.get('writtenaset'))
    with pytest.raises(PermissionError):
        co.columns
    with pytest.raises(PermissionError):
        repr(co)
    assert co.__dict__ == {}


def test_checkout_branch_not_existing_does_not_hold_writer_lock(two_commit_filled_samples_repo):
    repo = two_commit_filled_samples_repo
    assert 'doesnotexist' not in repo.list_branches()
    assert repo.writer_lock_held is False
    with pytest.raises(ValueError):
        co = repo.checkout(write=True, branch='doesnotexist')
    assert repo.writer_lock_held is False
    with pytest.raises(NameError):
        co.branch_name  # should not even exist
